# -*- coding: utf-8 -*-

import numpy as np
import matplotlib.pyplot as plt

def trapezoid(f, a, b, N):
    """Approximiert das Integral von `f` mit der Trapezregel.

    Input:
           f : Funktion f(x) welche integiert werden soll.
        a, b : untere/obere Grenze des Integrals.
           N : Anzahl Teilintervalle in der zusammengesetzten
               Trapezregel.
    """
    x, h = np.linspace(a, b, N+1, retstep=True)

    # Für geschlossene Quadraturformeln kann man mit einem
    # Trick die Anzahl Funktionsaufrufe reduzieren.
    # Die Beobachtung ist, dass zwei benachbarte Teilintervalle einen
    # gemeinsamen Knoten haben. Anstatt, `f` zweimal für diesen Knoten
    # auszurechnen, summiert man einfach die Gewichte:
    I = h * (np.sum(f(x[1:-1])) + 0.5 * (f(x[0]) + f(x[-1])))
    return I

def simpson(f, a, b, N):
    """Approximiert das Integral von `f` mit der Simpsonregel.

    Input:
           f : Funktion f(x) welche integiert werden soll.
        a, b : untere/obere Grenze des Integrals.
           N : Anzahl Teilintervalle in der zusammengesetzten
               Simpsonregel.
    """
    x, h = np.linspace(a, b, N+1, retstep=True)
    xm = 0.5*(x[1:] + x[:-1])

    I = h/6.0 * (4.0*np.sum(f(xm)) + 2.0*np.sum(f(x[1:-1])) + f(x[0]) + f(x[-1]))
    return I

def golub_welsch(n):
    """Berechnet die Knoten und Gewichte für Gauss-Legendre Quadratur.
    """

    x = np.zeros(n)
    w = np.zeros(n)

    # TODO implementieren Sie den Golub-Welsch Algorithmus.

    return x, w

def gauss_legendre(f, a, b, n):
    """ Gauss-Legendre Quadratur (nicht zusammengesetzt).

    f:     Funktion f(x)
    a, b:  Obere und untere Grenze des Intervalls.
    n:     Anzahl Quadraturpunkte.
    """

    # TODO implementieren Sie Gauss-Legendre

    return 0.0

def composite_legendre(f, a, b, N, n):
    """ Zusammengesetzte Gauss-Legendre Quadratur.

    f:     Funktion f(x)
    a, b:  Obere und untere Grenze des Intervalls.
    N:     Anzahl Teilintervalle.
    n:     Anzahl Quadraturpunkte pro Teilintervall.
    """

    # TODO implementieren Sie die zusammengesetzte Gauss-Legendre Quadratur.

    return 0.0

def quadrature_error(quadrature_rule, f, exact):
    """Berechnet den Fehler und die Konvergenzrate.

    Input:
        quadrature_rule : Quadraturformel für welche verwendet werden soll. Die
                          Funktion hat die gleiche Signature wie `simpson` und
                          `trapezoid`.

                      f : Funktion welche integriert werden soll. In der Aufgabe
                          heissen sie `f1` und `f2`.

                  exact : Exakte Referenzlösung des Integrals.

    Output:
                 errors : Fehler der Quadraturformel mit `n_chunks` Teilintervallen.
               n_chunks : Anzahl Teilintervalle.
    """

    n_chunks = np.arange(1, 101)
    errors = np.empty(n_chunks.shape)  # Vorsicht, n_chunks ist ein Array
                                       # voller `int`, i.e. `n_chunks.dtype`
                                       # ist `int`. Deswegen kein `empty_like`,
                                       # welches auch ein `int`-array wäre.

    for i, N in enumerate(n_chunks):
        approx = quadrature_rule(f, 0.0,  1.0, N)
        errors[i] = np.abs(approx - exact)

    return errors, n_chunks

def legendre_error(f, exact):
    # TODO Vorsichts halber gibt es eine eigene Funktion um den Fehler
    # von Gauss-Legendre zu berechnen. Denn Konvergenz ist in der Anzahl
    # Quadraturpunkte nicht Teilintervalle.
    # Damit die Anzahl Funktionsaufrufe vergleichbar ist, soll n = 1,...,100
    # gewählt werden.

    # Kann `compute_error` trotzdem verwendet werden?

    return [1.0], [0]

def composite_legendre_error(f, exact):
    # TODO Zusammengesetzte Gauss-Legendre Fehlerberechnung kann man auf
    # `quadrature_error` reduzieren (oder selber neu implementieren).

    # Die Anzahl Quadraturpunkte pro Intervall soll n = 4 sein, und die Anzahl
    # Teilintervalle N = 1,...,100.
    n = 4

    return [1.0], [0]

def plot_convergence(n_evals, errors, labels, title):
    """Plottet einen Konvergenzplot in einem neuen Fenster.

    Der Vergleich zwischen Simpson- und Trapezregel soll 'fair' sein. Die
    Simpsonregel wertet `f` fast 2x öfter aus. Plotten Sie deswegen den Fehler
    gegen die Anzahl Funktionsaufrufe.

    Trendlinien der Ordnung 1.5, 2 und 4 werden ebenfalls geplottet.

    Input:
        n_evals : list of array
                  Anzahl Funktionsaufrufe pro Regel pro Auflösung.

         errors : list of array
                  Fehler pro Quadraturregel pro Auflösung.

         labels : list. Text der in der Legende erscheint.
          title : Titel des Plots.
    """

    plt.figure(figsize=(12,8))

    for k, n_eval in enumerate(n_evals):
        plt.loglog(n_eval, errors[k], "-o", label=labels[k])

    # Trendlinien
    plt.loglog(n_evals[0], (1.0*n_evals[0])**-1.5, ":k", label=r"$n^{-1.5}$")
    plt.loglog(n_evals[0], (1.0*n_evals[0])**-2, "-k", label=r"$n^{-2}$")
    plt.loglog(n_evals[0], (1.0*n_evals[0])**-4, "-.k", label=r"$n^{-4}$")

    plt.grid(True)
    plt.xlabel("Anzahl Auswertungen von $f$")
    plt.ylabel("Absoluter Fehler")
    plt.legend(loc="lower left")
    plt.title(title)

def f1(x):
    return 1.0 / (1.0 + 5.0*x**2)

def f2(x):
    return np.sqrt(x)

def convergence_experiment(f, exact, title, filename):
    """Führt ein Konvergenzexperiment aus.

    Dies ist wieder eine der Funktionen welche das Problem: "Wiederholen Sie X
    für f = f1, f2, ..., fn." löst.

    Input:
               f : function. Entweder f1 oder f2.
           exact : Exakte Lösung.
           title : Titel des Plots.
        filename : plot wird hier gespeichert.

    """
    errors_tr, n_chunks = quadrature_error(trapezoid, f, exact)
    n_tr = n_chunks + 1
    errors_si, n_chunks = quadrature_error(simpson, f, exact)
    n_si = 2*n_chunks + 1

    errors_gl, n_gl = legendre_error(f, exact)
    errors_cg, n_cg = composite_legendre_error(f, exact)

    n_evals = [n_tr, n_si, n_gl, n_cg]
    errors = [errors_tr, errors_si, errors_gl, errors_cg]
    labels = ["Trapez", "Simpson", "Gauss-Legendre", "Zusammengesetzte Gauss Legende"]

    plot_convergence(n_evals, errors, labels, title)
    plt.savefig(filename + ".png")
    plt.savefig(filename + ".pdf")


def ex1_c():
    # Konvergenzplot für `f1`.
    I1ex = np.arctan(np.sqrt(5.0)) / np.sqrt(5.0)
    title = r"Quadraturfehler für $f_1(x) = \frac{1}{1 + 5x^2}$"
    convergence_experiment(f1, I1ex, title, "img/convergence_f1")

    # Konvergenzplot für `f2`.
    I2ex = 2.0 / 3.0
    title = r"Quadraturfehler für $f_2(x) = \sqrt{x}$"
    convergence_experiment(f2, I2ex, title, "img/convergence_f2")

if __name__ == "__main__":
    ex1_c()
    plt.show()
